package com.vernaio.example;

import java.io.File;
import java.util.List;
import java.util.stream.Collectors;

import com.vernaio.aivis.engine.v2.base.sdk.EngineLogger;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoFloatDataPoint;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.SignalPredictionData;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.SignalPredictionInference;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.SignalPredictionSetup;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.SignalPredictionTraining;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.dto.DtoInferenceConfig;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.dto.DtoTargetConfig;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.dto.DtoTrainingConfig;

public class Main {
	
	private static final int MIN_LOGGING_LEVEL = 3;
	
	private static final int MAX_THREADS = 0; // max threads 0 means the number of threads will be chosen automatically
	
	private static final String TRAINING_DATA_FILE = "train_sp.csv";
	
	private static final String INFERENCE_DATA_FILE = "eval_sp.csv";
	
	private static final String TARGET_SIGNAL_ID = "TARGET";
	
	public static void main(final String[] args) throws Exception {
		
		System.out.println("Setup ...");
		// parsing args as folder config
		final FolderConfig folderConfig = FolderConfig.fromArgs(args);
		// create and register logger
		SignalPredictionSetup.registerLogger(new EngineLogger() {
			
			@Override
			public void log(int level, String thread, String module, String message) {
				if (level <= MIN_LOGGING_LEVEL) {
					System.out.println(String.format("\t... %s", message));
				}
			}
		});
		// init thread count
		SignalPredictionSetup.initThreadCount(MAX_THREADS);
		System.out.println("\t... done!");
		
		/*
		 * TRAINING
		 */
		
		System.out.println("Loading training data ...");
		final SignalPredictionData trainingData = Utils.readCsvConvertToAivisFormat(new File(folderConfig.getInputFolder(), TRAINING_DATA_FILE));
		
		System.out.println("Creating training ...");
		// build training config - for more options here, see documentation
		final DtoTrainingConfig trainingConfig = new DtoTrainingConfig(new DtoTargetConfig(TARGET_SIGNAL_ID));
		
		// create training and train the model
		final SignalPredictionTraining training = SignalPredictionTraining.create(trainingData, trainingConfig);
		System.out.println("\t... done!");
		
		System.out.println("Retrieving and saving training files ...");
		// model is needed for inference and also contains the inference data specification
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "model.json"), training.getModel());
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "training-report.json"), training.getReport());
		System.out.println("\t... done!");
		
		/*
		 * INFERENCE
		 */
		
		// saving memory, we don't need training data anymore
		trainingData.destroy();
		System.out.println("Loading inference data ...");
		// it would also be enough to only read the signals which are part of the model's inference data spec - but we read all
		final SignalPredictionData inferenceData = Utils.readCsvConvertToAivisFormat(new File(folderConfig.getInputFolder(), INFERENCE_DATA_FILE));
		
		System.out.println("Creating inference ...");
		// build inference config - for more options here, see documentation
		// skipOnInsufficientData: if inference data does not meet the inference data specification for a timestamp an inference is demanded ("insufficient data"),
		// no exception is thrown but it's simply "skipped"
		final DtoInferenceConfig inferenceConfig = new DtoInferenceConfig(true);
		// note that we use here try-with-resource notation to destroy inference handle, inference.destroy() works equivalently
		try (final SignalPredictionInference inference = SignalPredictionInference.createByTraining(training, inferenceConfig)) {
			
			training.destroy();
			System.out.println("\t... done!");
			
			System.out.println("Calculating scores ...");
			// here, we choose exactly the timestamp where a target exists to get nice target-prediction-pairs to calculate kpis at
			final List<? extends IDtoFloatDataPoint> target = inferenceData.getFloatSignal(TARGET_SIGNAL_ID);
			final List<Long> inferenceTimestamps = target.stream().map(t -> t.getTimestamp()).collect(Collectors.toList());
			final List<? extends IDtoFloatDataPoint> predictions = inference.inferFloat(inferenceData, inferenceTimestamps);
			System.out.println("\t... done!");
			
			System.out.println("Saving inference files ...");
			Utils.writePredictions(new File(folderConfig.getOutputFolder(), "predictions.json"), predictions);
			Utils.writeAdditionalSignalsTimeseriesWithTarget(new File(folderConfig.getOutputFolder(), "additional-signals.json"), target, TARGET_SIGNAL_ID);
			inferenceData.destroy();
			System.out.println("\t... done!");
			
			// get list of predictions and timestamps where prediction exists
			// note predictions might not exist at all timestamps where target exists due to missing predecessor in features
			final List<Double> valuesPred = predictions.stream().map(IDtoFloatDataPoint::getValue).collect(Collectors.toList());
			final List<Long> timestampsPred = predictions.stream().map(IDtoFloatDataPoint::getTimestamp).collect(Collectors.toList());
			
			// get target values at timestampsPred
			final List<? extends IDtoFloatDataPoint> trueLabels = target.stream().filter(p -> timestampsPred.contains(p.getTimestamp())).collect(Collectors.toList());
			final List<Double> trueLabelValues = trueLabels.stream().map(p -> p.getValue()).collect(Collectors.toList());
			
			// print R^2 value of inference
			System.out.println("-----------------------------------------");
			System.out.printf("R^2 value of inference: " + Utils.computeR2(trueLabelValues, valuesPred) + "\n");
			System.out.println("-----------------------------------------");
		}
		
	}
}
