# this python script performs the same calculation as example_sp.ipynb

import json
import csv
import argparse
import os

from aivis_engine_v2_sp_sdk_python import (
    SignalPredictionSetup,
    SignalPredictionData,
    SignalPredictionTraining,
    SignalPredictionInference,
    EngineLogger,
    DtoFloatDataPoint,
)

# ---------------------------------------------------------------------------------
# Data location

TRAINING_DATA_FILE = "train_sp.csv"
INFERENCE_DATA_FILE = "eval_sp.csv"
TARGET_SIGNAL_ID = "TARGET"


# Input and output folder specified via args
def from_args():
    parser = argparse.ArgumentParser(
        description="Example project for signal prediction"
    )
    # data is expected in input folder
    parser.add_argument(
        "--input", "-i", type=str, nargs=1, required=True, help="Path to input folder"
    )
    # output files will be written to output folder
    parser.add_argument(
        "--output", "-o", type=str, nargs=1, required=True, help="Path to output folder"
    )
    args = parser.parse_args()
    # create output folder if not existing
    if not os.path.exists(args.output[0]):
        os.mkdir(args.output[0])
    return {
        "input_folder": args.input[0],
        "output_folder": args.output[0],
    }


# ---------------------------------------------------------------------------------
# Logger and Threads


class Logger(EngineLogger):
    def log(self, level, thread, module, message):
        min_logging_level = 3
        if level <= min_logging_level:
            print("\t... %s" % message)


def setup_logger_and_threads():
    print("Initializing SignalPredictionSetup: logging and threads ...")
    SignalPredictionSetup.register_logger(Logger())
    max_threads = 0  # number of threads will be chosen automatically with 0
    SignalPredictionSetup.init_thread_count(max_threads)
    print("\t... done!")
    return


# ---------------------------------------------------------------------------------
# Reading and Writing


def read_json(filename):
    print("\t... reading file '%s' ..." % filename)
    with open(filename, "r") as file:
        return json.loads(file.read())


def write_json(filename, obj):
    print("\t... writing file '%s' ..." % filename)
    with open(filename, "w") as file:
        file.write(json.dumps(obj, indent=2))
    return


# convert [DtoFloatDataPoint] to json
def float_data_points_to_timeseries_json(list_object):
    return {
        "_type": "Float",
        "dataPoints": [
            {
                "timestamp": int(point_object.timestamp),
                "value": float(point_object.value),
            }
            for point_object in list_object
        ],
    }


CHUNK_SIZE = 1024


# Main data reading function
# Float signals are added to an aivis data context via their id and a list of DtoFloatDataPoint.
# (Analogously for boolean and string signals.)
# The rest of the following function can be adapted if the data comes in a different format as in this example, e.g. in parquet files.
def read_csv_convert_to_aivis_format(filename):
    print("\t... reading file '%s' ..." % filename)

    # create empty data context for training data
    data_context = SignalPredictionData.create()

    # get line numbers to show nice progress bar
    line_count = sum(1 for _ in open(filename))

    # Read data in chunks to keep total memory usage low
    def consume_datapoints(signals):
        for signal in signals:
            data_context.add_float_signal(signal["signal_id"], signal["dataPoints"])
            signal["dataPoints"].clear()

    # read file
    with open(filename, newline="\n") as csvfile:
        # in this example, signal ids are in the header, first column is timestamps
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')
        headers = next(reader)
        signals = list(map(lambda h: {"signal_id": h, "dataPoints": []}, headers[1:]))

        # show nice progress bar
        progress_bar = -1
        line = 0

        nchunks = 0

        for n, row in enumerate(reader, start=1):
            # show nice progress bar
            progess_bar_new = round(line / (line_count / 10)) * 10
            if progress_bar != progess_bar_new:
                progress_bar = progess_bar_new
                print("\r\t... processing signals: %d%% ..." % progress_bar, end="")
            line += 1

            # get timestamp of this row, which is in the csv's first column
            timestamp = int(row[0])

            # add data points to all signal's list (if value exists)
            for i, column in enumerate(row[1:], start=0):
                if column:
                    # in this example, all signals are float signals
                    value = float(column)
                    dtofdp = DtoFloatDataPoint(timestamp, value)
                    signals[i]["dataPoints"].append(dtofdp)

            if n % CHUNK_SIZE == 0:
                nchunks += 1
                consume_datapoints(signals)

        print(
            f"\r\t... reading from file is done, processed {nchunks} chunks of {CHUNK_SIZE} rows!"
        )

        consume_datapoints(signals)  # add remaining signal values

        print("\r\t... data context is ready!")

    return data_context


# ---------------------------------------------------------------------------------
# MAIN


def main():
    folder_config = from_args()  # parsing args as folder_config
    setup_logger_and_threads()

    print("Loading training data ...")
    training_data = read_csv_convert_to_aivis_format(
        os.path.join(folder_config["input_folder"], TRAINING_DATA_FILE)
    )

    # print some signal to illustrate type DtoFloatDataPoint
    print("Example signal!")
    some_signal_id = "SIGNAL_7"
    print("\t... Signal id:", some_signal_id)
    some_signal_data_points = training_data.get_float_signal(some_signal_id)
    print("\t... Signal length:", len(some_signal_data_points))
    print("\t... Signal entry number 42:", some_signal_data_points[42])
    print(
        "\t... Signal entry number 42 timestamp (unix/millisec):",
        some_signal_data_points[42].timestamp,
    )
    print("\t... Signal entry number 42 value:", some_signal_data_points[42].value)

    # ---- TRAINING ----

    print("Creating training ...")
    # build training config - for more options see documentation
    training_config = json.dumps(
        {
            "target": {"signal": TARGET_SIGNAL_ID},
        }
    )

    # create training and train the model
    training = SignalPredictionTraining.create(training_data, training_config)
    print("\t... done!")

    print("Save training files ...")
    model = training.get_model()
    write_json(
        os.path.join(folder_config["output_folder"], "model.json"), json.loads(model)
    )

    training_report = json.loads(training.get_report())  # training infos
    write_json(
        os.path.join(folder_config["output_folder"], "training-report.json"),
        training_report,
    )
    print("\t... done!")

    # ---- INFERENCE ----
    print("Loading inference data ...")
    inference_data = read_csv_convert_to_aivis_format(
        os.path.join(folder_config["input_folder"], INFERENCE_DATA_FILE)
    )

    print("Creating inference ...")
    # build inference config
    inference_config = json.dumps({"skipOnInsufficientData": True})
    # via model is also an option: SignalPredictionInference.create_by_model(model_json, inference_config)
    inference = SignalPredictionInference.create_by_training(training, inference_config)
    print("\t... done!")

    print("Calculating predictions ...")
    # choose inference timestamps - exactly where the target has a data point to get comparable target-prediction-pairs for kpis
    # but of course any timestamp where the inference data meets the model's inference data specification is possible
    target = inference_data.get_float_signal(TARGET_SIGNAL_ID)
    timestamps = list(map(lambda p: p.timestamp, target))

    # infer predictions
    predictions = inference.infer_float(inference_data, timestamps)

    print("Save inference files ...")
    write_json(
        os.path.join(folder_config["output_folder"], "predictions.json"),
        float_data_points_to_timeseries_json(predictions),
    )
    write_json(
        os.path.join(folder_config["output_folder"], "additional-signals.json"),
        [
            {
                "id": TARGET_SIGNAL_ID,
                "timeseries": float_data_points_to_timeseries_json(target),
            }
        ],
    )
    print("\t... done!")

    print("Destroy everything ...")
    # destroying can be done once the respective object is not needed anymore - do this if memory is an issue
    training_data.destroy()
    training.destroy()
    inference_data.destroy()
    inference.destroy()
    print("\t... done!")
    # aivis part is

    # ---- KPI: compute r² value of inference----

    # get list of predictions and timestamps where predictions exist
    # note that predictions might not exist at all timestamps where target exists due to missing predecessor in features
    pred_values = list(map(lambda p: p.value, predictions))
    pred_timestamps = list(map(lambda p: p.timestamp, predictions))

    # get target values at timestamps_pred
    true_target = list(filter(lambda p: p.timestamp in pred_timestamps, target))
    true_target_values = list(map(lambda p: p.value, true_target))

    true_avg = sum(true_target_values) / len(true_target_values)
    rss = sum(
        [(true - pred) ** 2 for (true, pred) in zip(true_target_values, pred_values)]
    )
    tss = sum([(true - true_avg) ** 2 for true in true_target_values])

    print("-----------------------------------------")
    print("R² value of inference: ", 1 - rss / tss)
    print("-----------------------------------------")

    # ---- Toolbox HTML Report ----
    from aivis_engine_v2_toolbox.api import build_signal_prediction_report

    build_signal_prediction_report(
        {"title": "Aivis Engine V2 - Signal Prediction Example: Paper Quality",
         "predictions": {"file": os.path.join(folder_config["output_folder"], "predictions.json")},
         "target": {"file": os.path.join(folder_config["output_folder"], "additional-signals.json")},
         "trainingReportFile": os.path.join(folder_config["output_folder"], "training-report.json"),
         "outputFile": os.path.join(folder_config["output_folder"], "aivis-engine-v2-signal-prediction-example.html")})

    print("ALL DONE!")
    return


# run main function
main()
